/*
Copyright (c) 2003 Jasper van de Gronde. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
   3. The name of the author may not be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef BUFFERLIB_BUFFER_HPP
#define BUFFERLIB_BUFFER_HPP

//--------------------------------------------------------------------------- : Includes

#include "bufferlibtypes.hpp"
#include <vector>
#include <set>
#include <stdexcept>
#include <string>
#include <boost/shared_ptr.hpp>

//--------------------------------------------------------------------------- : Namespace

namespace bufferlib
{

//--------------------------------------------------------------------------- : buffer_t

template<typename T>
class input_t;
template<typename T>
class output_t;

// Base class for all buffer types
template<typename T>
class buffer_t
{
protected:
    std::vector<boost::shared_ptr<input_t<T> > > inputs_;
    std::vector<boost::shared_ptr<output_t<T> > > outputs_;
    std::multiset<count_t> input_positions_;
    std::multiset<count_t> output_positions_;

public:
    void add_input(const boost::shared_ptr<input_t<T> > input_ptr)
    {
        if ( inputs_.size() == 0 )
        {
            inputs_.push_back(input_ptr);
            input_positions_.insert(input_ptr->position());
        }
        else
        {
            throw std::runtime_error("Buffer already has an input!");
        }
    }
    void add_output(const boost::shared_ptr<output_t<T> > output_ptr)
    {
        outputs_.push_back(output_ptr);
        output_positions_.insert(output_ptr->position());
    }
    void remove_input(const boost::shared_ptr<input_t<T> > input_ptr)
    {
        const std::multiset<count_t>::const_iterator it = input_positions_.find(input_ptr->position());
        if ( it != input_positions_.end() )
        {
            input_positions_.erase(it); // Can't use erase directly, it would erase ALL entries equal to old_position.
        }
        inputs_.erase(input_ptr);
    }
    void remove_output(const boost::shared_ptr<input_t<T> > output_ptr)
    {
        const std::multiset<count_t>::const_iterator it = output_positions_.find(output_ptr->position());
        if ( it != output_positions_.end() )
        {
            output_positions_.erase(it); // Can't use erase directly, it would erase ALL entries equal to old_position.
        }
        outputs_.erase(output_ptr);
    }

protected:
    void update_input_position(count_t& position, const count_t count)
    {
        const std::multiset<count_t>::const_iterator it = input_positions_.find(position);
        if ( it != input_positions_.end() )
        {
            input_positions_.erase(it); // Can't use erase directly, it would erase ALL entries equal to old_position.
            input_positions_.insert(position+=count);
        }
        else
        {
            throw std::invalid_argument("Position doesn't exist in input_positions_!");
        }
    }
    void update_output_position(count_t& position, const count_t count)
    {
        const std::multiset<count_t>::const_iterator it = output_positions_.find(position);
        if ( it != output_positions_.end() )
        {
            output_positions_.erase(it); // Can't use erase directly, it would erase ALL entries equal to old_position.
            output_positions_.insert(position+=count);
        }
        else
        {
            throw std::invalid_argument("Position doesn't exist in output_positions_!");
        }
    }

public:
    count_t lowest_input_position() const
    {
        if ( !input_positions_.empty() )
        {
            return *(input_positions_.begin());
        }
        else
        {
            throw std::runtime_error("No input positions!");
        }
    }
    count_t lowest_output_position() const
    {
        if ( !output_positions_.empty() )
        {
            return *(output_positions_.begin());
        }
        else
        {
            throw std::runtime_error("No output positions!");
        }
    }

public:
    /*! Updates position if necessary! Should only be called by input_t. */
    virtual count_t write(const T* const, const count_t, count_t&) = 0;
    /*! Updates position if necessary! Should only be called by output_t. */
    virtual count_t read(T* const, const count_t, count_t&) = 0;

    virtual count_t can_write_from(const count_t) const = 0;
    virtual count_t can_read_from(const count_t) const = 0;
    virtual count_t can_write_min() const
    {
        if ( input_positions_.empty() )
        {
            return 0;
        }
        else
        {
            count_t min_count=can_write_from(*(input_positions_.begin()));
            for(std::multiset<count_t>::const_iterator it=++(input_positions_.begin()); it!=input_positions_.end(); ++it)
            {
                const count_t count = can_write_from(*it);
                if ( count == 0 )
                {
                    return 0;
                }
                else if ( count < min_count )
                {
                    min_count=count;
                }
            }
            return min_count;
        }
    }
    virtual count_t can_read_min() const
    {
        if ( output_positions_.empty() )
        {
            return 0;
        }
        else
        {
            count_t min_count=can_read_from(*(output_positions_.begin()));
            for(std::multiset<count_t>::const_iterator it=++(output_positions_.begin()); it!=output_positions_.end(); ++it)
            {
                const count_t count = can_read_from(*it);
                if ( count == 0 )
                {
                    return 0;
                }
                else if ( count < min_count )
                {
                    min_count=count;
                }
            }
            return min_count;
        }
    }

    virtual ~buffer_t()
    {}
}
;

//--------------------------------------------------------------------------- : EOF

} // Namespace bufferlib

#endif //BUFFERLIB_BUFFER_HPP
